from collections import abc
from itertools import islice
from typing import (Generic,
                    Iterator,
                    Tuple,
                    Union)

from dendroid import red_black
from dendroid.hints import Set as RawSet
from reprit.base import (generate_repr,
                         seekers)

from .core.tokenization import (Token,
                                Tokenizer)
from .hints import Domain

AnyNode = Union[red_black.NIL, red_black.Node]


@abc.MutableSet.register
class Set(Generic[Domain]):
    __slots__ = '_values', '_tokenizer'

    def __init__(self, *values: Domain) -> None:
        self._values = red_black.set_(*values)
        self._tokenizer = Tokenizer()

    __repr__ = generate_repr(__init__,
                             field_seeker=seekers.complex_)

    def __and__(self, other: 'Set[Domain]') -> 'Set[Domain]':
        return (self._from_raw(self._values & other._values)
                if isinstance(other, Set)
                else NotImplemented)

    def __bool__(self) -> bool:
        return bool(self._values)

    def __contains__(self, value: Domain):
        return value in self._values

    def __eq__(self, other: 'Set[Domain]') -> bool:
        return (self._values == other._values
                if isinstance(other, Set)
                else NotImplemented)

    def __iand__(self, other: 'Set[Domain]') -> 'Set[Domain]':
        if not isinstance(other, Set):
            return NotImplemented
        size = len(self._values)
        if len(other._values) != size:
            self._tokenizer.reset()
            self._values &= other._values
        else:
            common_values = self._values & other._values
            if len(common_values) != size:
                self._tokenizer.reset()
                self._values = common_values
        return self

    def __ior__(self, other: 'Set[Domain]') -> 'Set[Domain]':
        if not isinstance(other, Set):
            return NotImplemented
        extra_values = other._values - self._values
        if extra_values:
            self._tokenizer.reset()
            self._values |= extra_values
        return self

    def __isub__(self, other: 'Set[Domain]') -> 'Set[Domain]':
        if not isinstance(other, Set):
            return NotImplemented
        common_values = self._values & other._values
        if common_values:
            self._tokenizer.reset()
            self._values -= common_values
        return self

    def __iter__(self) -> 'SetForwardIterator[Domain]':
        return SetForwardIterator(0, self._values.tree.min(),
                                  self._values.tree, self._tokenizer.create())

    def __ixor__(self, other: 'Set[Domain]') -> 'Set[Domain]':
        if not isinstance(other, Set):
            return NotImplemented
        if other:
            self._tokenizer.reset()
            self._values ^= other._values
        return self

    def __len__(self) -> int:
        return len(self._values)

    def __lt__(self, other: 'Set[Domain]') -> bool:
        return (self._values < other._values
                if isinstance(other, Set)
                else NotImplemented)

    def __le__(self, other: 'Set[Domain]') -> bool:
        return (self._values <= other._values
                if isinstance(other, Set)
                else NotImplemented)

    def __or__(self, other: 'Set[Domain]') -> 'Set[Domain]':
        return (self._from_raw(self._values | other._values)
                if isinstance(other, Set)
                else NotImplemented)

    def __reversed__(self) -> 'SetBackwardIterator[Domain]':
        return SetBackwardIterator(0, self._values.tree.max(),
                                   self._values.tree, self._tokenizer.create())

    def __sub__(self, other: 'Set[Domain]') -> 'Set[Domain]':
        return (self._from_raw(self._values - other._values)
                if isinstance(other, Set)
                else NotImplemented)

    def __xor__(self, other: 'Set[Domain]') -> 'Set[Domain]':
        return (self._from_raw(self._values ^ other._values)
                if isinstance(other, Set)
                else NotImplemented)

    @classmethod
    def _from_raw(cls, raw: RawSet[Domain]) -> 'Set[Domain]':
        result = Set()
        result._values = raw
        return result

    def add(self, value: Domain) -> None:
        if value not in self._values:
            self._tokenizer.reset()
            self._values.add(value)

    def begin(self) -> 'SetForwardIterator[Domain]':
        return SetForwardIterator(0, self._values.tree.min(),
                                  self._values.tree, self._tokenizer.create())

    def clear(self) -> None:
        self._tokenizer.reset()
        self._values.clear()

    def discard(self, value: Domain) -> None:
        node = self._values.tree.find(value)
        if node is not red_black.NIL:
            self._tokenizer.reset()
            self._values.tree.remove(node)

    def end(self) -> 'SetForwardIterator[Domain]':
        return SetForwardIterator(len(self._values), red_black.NIL,
                                  self._values.tree, self._tokenizer.create())

    def isdisjoint(self, other: 'Set[Domain]') -> bool:
        return self._values.isdisjoint(other._values)

    def max(self) -> Domain:
        return self._values.max()

    def min(self) -> Domain:
        return self._values.min()

    def pop(self) -> Domain:
        if self._values:
            self._tokenizer.reset()
        return self._values.pop()

    def rbegin(self) -> 'SetBackwardIterator[Domain]':
        return SetBackwardIterator(0, self._values.tree.max(),
                                   self._values.tree, self._tokenizer.create())

    def remove(self, value: Domain) -> None:
        node = self._values.tree.find(value)
        if node is red_black.NIL:
            raise ValueError('{!r} is not found.'.format(value))
        else:
            self._tokenizer.reset()
            self._values.tree.remove(node)

    def rend(self) -> 'SetBackwardIterator[Domain]':
        return SetBackwardIterator(len(self._values), red_black.NIL,
                                   self._values.tree, self._tokenizer.create())


class SetBackwardIterator(Iterator[Domain]):
    __slots__ = '_index', '_node', '_tree', '_token'

    def __init__(self,
                 index: int,
                 node: AnyNode,
                 tree: red_black.Tree[Domain, Domain],
                 token: Token) -> None:
        self._index = index
        self._node = node
        self._tree = tree
        self._token = token

    def __add__(self, offset: int) -> 'SetBackwardIterator[Domain]':
        return SetBackwardIterator(*self._move_node(offset), self._tree,
                                   self._token)

    def __eq__(self, other: 'SetBackwardIterator[Domain]') -> bool:
        return (self._to_validated_node() is other._to_validated_node()
                if isinstance(other, SetBackwardIterator)
                else NotImplemented)

    def __iadd__(self, offset: int) -> 'SetBackwardIterator[Domain]':
        self._index, self._node = self._move_node(offset)
        return self

    def __isub__(self, offset: int) -> 'SetBackwardIterator[Domain]':
        self._index, self._node = self._move_node(-offset)
        return self

    def __iter__(self) -> 'SetBackwardIterator[Domain]':
        return self

    def __next__(self) -> Domain:
        node = self._to_validated_node()
        try:
            result = node.value
        except AttributeError:
            raise StopIteration from None
        else:
            self._index += 1
            self._node = self._tree.predecessor(self._node)
            return result

    def __sub__(self, offset: int) -> 'SetBackwardIterator[Domain]':
        return SetBackwardIterator(*self._move_node(-offset), self._tree,
                                   self._token)

    def _move_node(self, offset: int) -> Tuple[int, AnyNode]:
        self._validate()
        index = self._index
        size = len(self._tree)
        min_offset, max_offset = -index, size - index
        if offset < min_offset or offset > max_offset:
            raise ValueError('Offset should be '
                             'in range({min_offset}, {max_offset}), '
                             'but found {offset}.'
                             .format(min_offset=min_offset,
                                     max_offset=max_offset + 1,
                                     offset=offset)
                             if size
                             else 'Set is empty.')
        new_index = index + offset
        return new_index, (_index_to_node(size - new_index - 1, self._tree)
                           if new_index < size
                           else red_black.NIL)

    def _to_validated_node(self) -> AnyNode:
        self._validate()
        return self._node

    def _validate(self) -> None:
        if self._token.expired:
            raise ValueError('Iterator is invalidated.')


class SetForwardIterator(Iterator[Domain]):
    __slots__ = '_index', '_node', '_tree', '_token'

    def __init__(self,
                 index: int,
                 node: AnyNode,
                 tree: red_black.Tree[Domain, Domain],
                 token: Token) -> None:
        self._index = index
        self._node = node
        self._tree = tree
        self._token = token

    def __add__(self, offset: int) -> 'SetForwardIterator[Domain]':
        return SetForwardIterator(*self._move_node(offset), self._tree,
                                  self._token)

    def __eq__(self, other: 'SetForwardIterator[Domain]') -> bool:
        return (self._to_validated_node() is other._to_validated_node()
                if isinstance(other, SetForwardIterator)
                else NotImplemented)

    def __iadd__(self, offset: int) -> 'SetForwardIterator[Domain]':
        self._index, self._node = self._move_node(offset)
        return self

    def __isub__(self, offset: int) -> 'SetForwardIterator[Domain]':
        self._index, self._node = self._move_node(-offset)
        return self

    def __iter__(self) -> 'SetForwardIterator[Domain]':
        return self

    def __next__(self) -> Domain:
        node = self._to_validated_node()
        try:
            result = node.value
        except AttributeError:
            raise StopIteration from None
        else:
            self._node = self._tree.successor(self._node)
            return result

    def __sub__(self, offset: int) -> 'SetForwardIterator[Domain]':
        return SetForwardIterator(*self._move_node(-offset), self._tree,
                                  self._token)

    def _move_node(self, offset: int) -> Tuple[int, AnyNode]:
        self._validate()
        index = self._index
        size = len(self._tree)
        min_offset, max_offset = -index, size - index
        if offset < min_offset or offset > max_offset:
            raise ValueError('Offset should be '
                             'in range({min_offset}, {max_offset}), '
                             'but found {offset}.'
                             .format(min_offset=min_offset,
                                     max_offset=max_offset + 1,
                                     offset=offset)
                             if size
                             else 'Set is empty.')
        new_index = index + offset
        return new_index, (_index_to_node(new_index, self._tree)
                           if new_index < size
                           else red_black.NIL)

    def _to_validated_node(self) -> AnyNode:
        self._validate()
        return self._node

    def _validate(self) -> None:
        if self._token.expired:
            raise ValueError('Iterator is invalidated.')


def _node_to_index(node: red_black.Node, tree: red_black.Tree) -> int:
    return next(index
                for index, candidate in enumerate(tree)
                if candidate is node)


def _index_to_node(index: int, tree: red_black.Tree) -> red_black.Node:
    return next(islice(tree, index, None))



import sys
from collections import abc
from typing import (Generic,
                    Iterable,
                    Iterator,
                    List,
                    Optional,
                    Union,
                    overload)

from reprit.base import (generate_repr,
                         seekers)

from .core.tokenization import (Token,
                                Tokenizer)
from .hints import Domain


@abc.MutableSequence.register
class Vector(Generic[Domain]):
    __slots__ = '_values', '_tokenizer'

    def __init__(self, *values: Domain) -> None:
        self._values = list(values)  # type: List[Domain]
        self._tokenizer = Tokenizer()

    __repr__ = generate_repr(__init__,
                             field_seeker=seekers.complex_)

    def __add__(self, other: 'Vector[Domain]') -> 'Vector[Domain]':
        return (Vector(*self._values, *other._values)
                if isinstance(other, Vector)
                else NotImplemented)

    def __bool__(self) -> bool:
        return bool(self._values)

    def __contains__(self, value: Domain) -> bool:
        return value in self._values

    def __delitem__(self, item: Union[int, slice]) -> None:
        del self._values[item]

    def __eq__(self, other: 'Vector') -> bool:
        return (self._values == other._values
                if isinstance(other, Vector)
                else NotImplemented)

    @overload
    def __getitem__(self, item: int) -> Domain:
        """Returns element by given index."""

    @overload
    def __getitem__(self, item: slice) -> 'Vector[Domain]':
        """Returns subvector by given slice."""

    def __getitem__(self, item):
        return (self._values[item]
                if isinstance(item, int)
                else Vector(*self._values[item]))

    def __iadd__(self, values: Iterable[Domain]) -> 'Vector[Domain]':
        self.extend(values)
        return self

    def __iter__(self) -> 'VectorForwardIterator[Domain]':
        return VectorForwardIterator(0, self._values, self._tokenizer.create())

    def __len__(self) -> int:
        return len(self._values)

    def __le__(self, other: 'Vector') -> bool:
        return (self._values <= other._values
                if isinstance(other, Vector)
                else NotImplemented)

    def __lt__(self, other: 'Vector') -> bool:
        return (self._values < other._values
                if isinstance(other, Vector)
                else NotImplemented)

    def __reversed__(self) -> 'VectorBackwardIterator[Domain]':
        return VectorBackwardIterator(0, self._values,
                                      self._tokenizer.create())

    @overload
    def __setitem__(self, item: int, value: Domain) -> None:
        """Sets element by given index to given value."""

    @overload
    def __setitem__(self, item: slice, values: Iterable[Domain]) -> None:
        """Sets subvector by given slice to given values."""

    def __setitem__(self, item, value):
        if isinstance(item, slice):
            start, stop, step = item.indices(len(self._values))
            value = list(value)
            if stop > start or step == 1 and value:
                self._tokenizer.reset()
        self._values[item] = value

    def append(self, value: Domain) -> None:
        self._tokenizer.reset()
        self._values.append(value)

    def begin(self) -> 'VectorForwardIterator[Domain]':
        return VectorForwardIterator(0, self._values,
                                     self._tokenizer.create())

    def clear(self) -> None:
        self._tokenizer.reset()
        self._values.clear()

    def count(self, value: Domain) -> int:
        return self._values.count(value)

    def end(self) -> 'VectorForwardIterator[Domain]':
        return VectorForwardIterator(len(self), self._values,
                                     self._tokenizer.create())

    def extend(self, values: Iterable[Domain]) -> None:
        iterator = iter(values)
        try:
            value = next(iterator)
        except StopIteration:
            return
        else:
            self._tokenizer.reset()
        self._values.append(value)
        self._values.extend(iterator)

    def index(self,
              value: Domain,
              start: int = 0,
              stop: int = sys.maxsize) -> int:
        return self._values.index(value, start, stop)

    def insert(self, index: int, value: Domain) -> None:
        self._tokenizer.reset()
        self._values.insert(index, value)

    def pop(self, index: int = -1) -> None:
        if self._values:
            self._tokenizer.reset()
        return self._values.pop(index)

    def pop_back(self) -> None:
        self._tokenizer.reset()
        del self._values[-1]

    push_back = append

    def rbegin(self) -> 'VectorBackwardIterator[Domain]':
        return VectorBackwardIterator(0, self._values,
                                      self._tokenizer.create())

    def remove(self, value: Domain) -> None:
        try:
            index = self._values.index(value)
        except ValueError:
            raise
        else:
            self._tokenizer.reset()
            del self._values[index]

    def rend(self) -> 'VectorBackwardIterator[Domain]':
        return VectorBackwardIterator(len(self), self._values,
                                      self._tokenizer.create())

    def resize(self, size: int, value: Optional[Domain] = None) -> None:
        if size < 0:
            raise ValueError('Size should be positive, but found {}.'
                             .format(size))
        self._tokenizer.reset()
        self._values = (self._values
                        + [value] * (size - len(self._values)))[:size]

    def reverse(self) -> None:
        self._tokenizer.reset()
        self._values.reverse()


class VectorBackwardIterator(Iterator[Domain]):
    __slots__ = '_index', '_values', '_token'

    def __init__(self, index: int, values: List[Domain], token: Token) -> None:
        self._index = index
        self._values = values
        self._token = token

    def __add__(self, offset: int) -> 'VectorBackwardIterator[Domain]':
        return VectorBackwardIterator(self._move_index(offset), self._values,
                                      self._token)

    def __eq__(self, other: 'VectorBackwardIterator[Domain]') -> bool:
        return (self._to_validated_values() is other._to_validated_values()
                and self._index == other._index
                if isinstance(other, VectorBackwardIterator)
                else NotImplemented)

    def __iadd__(self, offset: int) -> 'VectorBackwardIterator[Domain]':
        self._index = self._move_index(offset)
        return self

    def __isub__(self, offset: int) -> 'VectorBackwardIterator[Domain]':
        self._index = self._move_index(-offset)
        return self

    def __iter__(self) -> 'VectorBackwardIterator[Domain]':
        return self

    def __le__(self, other: 'VectorBackwardIterator[Domain]') -> bool:
        return (self._to_validated_values() is other._to_validated_values()
                and self._index <= other._index
                if isinstance(other, VectorBackwardIterator)
                else NotImplemented)

    def __lt__(self, other: 'VectorBackwardIterator[Domain]') -> bool:
        return (self._to_validated_values() is other._to_validated_values()
                and self._index < other._index
                if isinstance(other, VectorBackwardIterator)
                else NotImplemented)

    def __next__(self) -> Domain:
        values = self._to_validated_values()
        try:
            result = values[-self._index - 1]
        except IndexError:
            raise StopIteration from None
        else:
            self._index += 1
            return result

    def __sub__(self, offset: int) -> 'VectorBackwardIterator[Domain]':
        return VectorBackwardIterator(self._move_index(-offset), self._values,
                                      self._token)

    def _move_index(self, offset: int) -> int:
        size = len(self._to_validated_values())
        min_offset, max_offset = -self._index, size - self._index
        if offset < min_offset or offset > max_offset:
            raise ValueError('Offset should be '
                             'in range({min_offset}, {max_offset}), '
                             'but found {offset}.'
                             .format(min_offset=min_offset,
                                     max_offset=max_offset + 1,
                                     offset=offset)
                             if size
                             else 'Vector is empty.')
        return self._index + offset

    def _to_validated_values(self) -> List[Domain]:
        self._validate()
        return self._values

    def _validate(self) -> None:
        if self._token.expired:
            raise ValueError('Iterator is invalidated.')


class VectorForwardIterator(Iterator[Domain]):
    __slots__ = '_index', '_values', '_token'

    def __init__(self, index: int, values: List[Domain], token: Token) -> None:
        self._index = index
        self._values = values
        self._token = token

    def __add__(self, offset: int) -> 'VectorForwardIterator[Domain]':
        return VectorForwardIterator(self._move_index(offset), self._values,
                                     self._token)

    def __eq__(self, other: 'VectorForwardIterator[Domain]') -> bool:
        return (self._to_validated_values() is other._to_validated_values()
                and self._index == other._index
                if isinstance(other, VectorForwardIterator)
                else NotImplemented)

    def __iadd__(self, offset: int) -> 'VectorForwardIterator[Domain]':
        self._index = self._move_index(offset)
        return self

    def __isub__(self, offset: int) -> 'VectorForwardIterator[Domain]':
        self._index = self._move_index(-offset)
        return self

    def __iter__(self) -> 'VectorForwardIterator[Domain]':
        return self

    def __le__(self, other: 'VectorForwardIterator[Domain]') -> bool:
        return (self._to_validated_values() is other._to_validated_values()
                and self._index <= other._index
                if isinstance(other, VectorForwardIterator)
                else NotImplemented)

    def __lt__(self, other: 'VectorForwardIterator[Domain]') -> bool:
        return (self._to_validated_values() is other._to_validated_values()
                and self._index < other._index
                if isinstance(other, VectorForwardIterator)
                else NotImplemented)

    def __next__(self) -> Domain:
        values = self._to_validated_values()
        try:
            result = values[self._index]
        except IndexError:
            raise StopIteration from None
        else:
            self._index += 1
            return result

    def __sub__(self, offset: int) -> 'VectorForwardIterator[Domain]':
        return VectorForwardIterator(self._move_index(-offset), self._values,
                                     self._token)

    def _move_index(self, offset: int) -> int:
        size = len(self._to_validated_values())
        min_offset, max_offset = -self._index, size - self._index
        if offset < min_offset or offset > max_offset:
            raise ValueError('Offset should be '
                             'in range({min_offset}, {max_offset}), '
                             'but found {offset}.'
                             .format(min_offset=min_offset,
                                     max_offset=max_offset + 1,
                                     offset=offset)
                             if size
                             else 'Vector is empty.')
        return self._index + offset

    def _to_validated_values(self) -> List[Domain]:
        self._validate()
        return self._values

    def _validate(self) -> None:
        if self._token.expired:
            raise ValueError('Iterator is invalidated.')



import weakref

try:
    from typing import Protocol
except ImportError:
    from typing_extensions import Protocol


class WeakReferencable(Protocol):
    __weakref__ = ...


class Token:
    __slots__ = '_value',

    def __init__(self, value: WeakReferencable) -> None:
        self._value = weakref.ref(value)

    @property
    def expired(self) -> bool:
        return self._value() is None


class Tokenizer:
    class _ValueFactory:
        __slots__ = '__weakref__',

    __slots__ = '_value',

    def __init__(self) -> None:
        self._value = self._ValueFactory()  # type: WeakReferencable

    def create(self) -> Token:
        return Token(self._value)

    def reset(self) -> None:
        del self._value
        self._value = self._ValueFactory()
